#!/usr/bin/env python2.6
# -*- coding: utf-8 -*-
# This software is distributed under the two-clause BSD license.

__version__ = '0.3.8'
__author__ = "Raphaël Barrois <raphael.barrois+uconf@polytechnique.org>"



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

"""Handles parsing of rules."""

import re

import tdparser


# {{{ Nodes

class _ConditionNode(object):
    """Base class for a node."""

    def eval(self, atoms):
        """Evaluate this node with a given set of atoms.

        Returns a boolean.
        """
        raise NotImplementedError()

    def simplify(self):
        """Simplify the current node (for easier representation).

        Returns an equivalent node.
        """
        return self

    def __eq__(self, other):
        if not isinstance(other, _ConditionNode):
            return NotImplemented

        return repr(self.simplify()) == repr(other.simplify())


class _FalseNode(_ConditionNode):
    """A 'false' node."""
    def eval(self, atoms):
        return False

    def __repr__(self):
        return '<False>'

    def __str__(self):
        return 'False'


class _TrueNode(_ConditionNode):
    """A 'true' node."""
    def eval(self, atoms):
        return True

    def __repr__(self):
        return '<True>'

    def __str__(self):
        return 'True'


class _TextNode(_ConditionNode):
    """A 'text' node.

    Attributes:
        text (str): the atom contained in the node.
    """

    def __init__(self, text):
        self.text = text

    def eval(self, atoms):
        return self.text in atoms

    def __repr__(self):
        return '<%s>' % self.text

    def __str__(self):
        return self.text


class _NegateNode(_ConditionNode):
    """A 'not' node.

    Attributes:
        son (_ConditionNode): the negated node.
    """

    precedence = 30

    def __init__(self, son):
        self.son = son

    def eval(self, atoms):
        return not self.son.eval(atoms)

    def simplify(self):
        self.son = self.son.simplify()
        return self

    def __repr__(self):
        return '<Not %s>' % self.son

    def __str__(self):
        return '! %s' % self.son


class _MultiNode(_ConditionNode):
    """An abstract node with many sons.

    Attributes:
        sons (iterable of _ConditionNode): the sons of this node.
    """

    def __init__(self, sons):
        self.sons = sons

    def simplify(self):
        """Merge all sons if possible."""

        if len(self.sons) == 1:
            # Only one son, return it (simplified)
            return self.sons[0].simplify()

        new_sons = []
        for son in self.sons:
            # Simplify sons
            son = son.simplify()
            if son.__class__ == self.__class__:
                # If they are instances of this node, merge their sons too
                new_sons.extend(son.sons)
            else:
                new_sons.append(son)
        self.sons = new_sons
        return self


class _AndNode(_MultiNode):
    """A 'and' node."""
    precedence = 20

    def eval(self, atoms):
        return all(son.eval(atoms) for son in self.sons)

    def __repr__(self):
        return '<And%r>' % (tuple(self.sons),)

    def __str__(self):
        return ' && '.join(
            ('(%s)' % son if isinstance(son, _MultiNode) else str(son))
            for son in self.sons)


class _OrNode(_MultiNode):
    """A 'or' node."""
    precedence = 10

    def eval(self, atoms):
        return any(son.eval(atoms) for son in self.sons)

    def __repr__(self):
        return '<Or%r>' % (tuple(self.sons),)

    def __str__(self):
        return ' || '.join(
            ('(%s)' % son if isinstance(son, _MultiNode) else str(son))
            for son in self.sons)


# }}}
# {{{ Tokens


class _TextToken(tdparser.Token):
    lbp = 25
    regexp = r'[a-zA-Z0-9._-]+'

    def __init__(self, text):
        self.text = text

    def nud(self, context):
        return _TextNode(self.text)

    def led(self, left, context):
        return _OrNode([left, _TextNode(self.text)])

    def __repr__(self):
        return '<Text: %r>' % self.text


class _OrToken(tdparser.Token):
    lbp = 10
    regexp = r'\|\|'

    def led(self, left, context):
        right = context.expression(self.lbp)
        return _OrNode([left, right])

    def __repr__(self):
        return '<Or>'

class _AndToken(tdparser.Token):
    lbp = 15
    regexp = r'&&'

    def led(self, left, context):
        right = context.expression(self.lbp)
        return _AndNode([left, right])

    def __repr__(self):
        return '<And>'


class _NotToken(tdparser.Token):
    lbp = 10
    regexp = r'!'

    def nud(self, context):
        return _NegateNode(context.expression(100))

    def __repr__(self):
        return '<Not>'


# }}}
# {{{ Lexer


class RuleLexer(object):
    """Build a rule lexer

    Some possible patterns:
    a b c => Matches if any of a, b, c
    a || b || c
    a || (b && !c) => Matches if a or (b and not c)
    """
    def __init__(self):
        self.lexer = self._build_lexer()

    @classmethod
    def _build_lexer(cls):
        lexer = tdparser.Lexer(with_parens=True)
        lexer.register_tokens(_TextToken, _OrToken, _AndToken, _NotToken)
        return lexer

    def get_rule(self, text):
        return Rule(text, self.lexer.parse(text))

# }}}
# {{{ Rule

class Rule(object):
    def __init__(self, text, node):
        self.text = text
        self.node = node

    def test(self, categories):
        """Test whether a set of categories match this rule.

        Args:
            categories (str set): categories to test

        Returns:
            bool
        """
        return self.node.eval(categories)

    def __repr__(self):
        return "Rule(%r)" % self.text

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.node == other.node

# }}}



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

"""Handles parsing of action options."""

import collections
import re
import tdparser


Option = collections.namedtuple('Option', ('key', 'value'))


class IdentifierToken(tdparser.Token):
    """An option."""
    lbp = 20
    regexp = r'[a-z0-9_-]+'

    def __init__(self, text):
        self.text = text

    def nud(self, context):
        """Try to retrieve the value."""
        return [Option(self.text, None)]


class EqualToken(tdparser.Token):
    """The equal sign, between an identifier and a token."""
    lbp = 30
    regexp = r'='

    def led(self, left, context):
        identifier, value = left[0]

        if value is not None:
            raise ValueError("Invalid chained '%s=%s=%s'." % (
                identifier, value, context.consume().text))

        if isinstance(context.current_token, IdentifierToken):
            # The lexer mistook our value for an identifier, let's take it.
            value = context.consume().text
        else:
            # That's a "standard" value, let's pick it.
            value = context.expression(self.lbp)

        return [Option(identifier, value)]


class SpaceToken(tdparser.Token):
    """A space (separates options).

    Lower binding power so that id=value gets glued together.
    """
    lbp = 10
    regexp = r' +'

    def nud(self, context):
        # At the beginning of the string
        return context.expression(self.lbp)

    def led(self, left, context):
        """Simply sum expressions on the left and on the right."""
        return left + context.expression(self.lbp)


class BaseValueToken(tdparser.Token):
    """Any valid value."""
    lbp = 20

    def clean_text(self):
        return self.text

    def nud(self, context):
        return self.clean_text()


class UnQuotedTextToken(BaseValueToken):
    regexp = r"[^'\" ]+"


class BaseQuotedTextToken(BaseValueToken):

    def clean_text(self):
        text = self.text[1:-1]
        # Strip a level of backslashes
        text = re.sub(r'\\(.)', r'\1', text)
        return text


class DoubleQuotedTextToken(BaseQuotedTextToken):
    regexp = r'"([^\\"]|\\.)*"'


class SingleQuotedTextToken(BaseQuotedTextToken):
    regexp = r"'([^\\']|\\.)*'"


class ActionLexer(object):
    def __init__(self, lexer=None):
        self.lexer = lexer or self._build_lexer()

    @classmethod
    def _build_lexer(cls):
        lexer = tdparser.Lexer(with_parens=False, blank_chars='')
        lexer.register_tokens(IdentifierToken, EqualToken, DoubleQuotedTextToken, SingleQuotedTextToken, SpaceToken)
        return lexer

    def get_options(self, text):
        options = self.lexer.parse(text.strip())
        return dict(options)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.


from __future__ import unicode_literals

import os.path
import sys

from confutils import Default

from . import __version__
from . import config
from . import fs
from . import helpers
from . import porcelain


class UConfError(Exception):
    pass


class ConfigError(UConfError):
    pass


class BaseCommand(object):
    """Base command object."""
    name = ''
    help = ''

    @classmethod
    def register_options(cls, parser):
        """Register command-specific options into an argparse subparser."""

    @classmethod
    def get_name(cls):
        """Retrieve the name of the command.

        Will try in order:
        - ``name`` attribute
        - lowercase class name, stripping the 'command' part.
        """
        if cls.name:
            return cls.name
        name = cls.__name__.lower()
        if name.endswith('command'):
            name = name[:-len('command')]
        return name

    @classmethod
    def get_help(cls):
        """Retrieve the help text for a command.

        If the ``help`` attribute is empty, use the class' docstring.
        """
        if cls.help:
            return cls.help
        else:
            return cls.__doc__

    required_config_fields = ()

    def __init__(self, env, parser):
        self.env = env
        self.parser = parser

        self.stdout = sys.stdout
        self.stderr = sys.stderr

        self.check_required_config()

    def check_required_config(self):
        for field in self.required_config_fields:
            if not self.env.isset(field):
                raise ConfigError("Field '%s' must be set, either in config files "
                    "or through command-line arguments." % field)

    def warning(self, message, *args):
        self.stderr.write(message % args)
        self.stderr.write('\n')
        if self.options.get('strict', False):
            self.stderr.write("Strict mode: aborting.\n")
            raise SystemExit(1)

    def info(self, message, *args):
        self.stdout.write(message % args)
        self.stdout.write('\n')

    def run(self):
        """Run the actual command."""
        pass


class HelpCommand(BaseCommand):
    name = 'help'
    help = "List of options and commands"

    def run(self):
        self.parser.print_help()


class VersionCommand(BaseCommand):
    name = 'version'
    help = "Display the current version number"

    def run(self):
        self.stdout.write('%(prog)s %(version)s\n' % dict(
            prog=self.parser.prog, version=__version__))


class Init(BaseCommand):
    name = 'init'
    help = "Setup a new uconf repository"

    required_config_fields = ('root',)

    def run(self):
        self.env.root = self.env.get('root')
        fs = self.env.get_repo_fs()
        fs.makedir(env.uconf_dir, recursive=True, allow_recreate=True)
        fs.writelines(os.path.join(env.uconf_dir, 'config'), [])


class WithRepoCommand(BaseCommand):
    """Enhanced base command class with list of rules already parsed."""

    def __init__(self, *args, **kwargs):
        super(WithRepoCommand, self).__init__(*args, **kwargs)

        initial_cats = self.env.getlist('initial', helpers.get_hostnames())
        self.active_repository = self.env.get_active_repository(initial_cats)

    def _get_files(self, files):
        """Retrieve file config for a set of file names.

        If no filename was provided, return all files.
        """
        all_files = self.active_repository.iter_files()

        return helpers.filter_iter(all_files, files, empty_is_all=True)


class Make(WithRepoCommand):
    """Make one or more files."""

    name = 'make'
    help = "Build and install one or more files."

    required_config_fields = ('target',)

    @classmethod
    def register_options(cls, parser):
        parser.add_argument('files', nargs='*', default=Default(tuple()),
            help="Build selected files, all valid if empty.")
        super(Make, cls).register_options(parser)

    def run(self):
        p = porcelain.MakeFile(self.env, self.active_repository)
        for filename in self._get_files(self.env.get('files')):
            try:
                p.handle(filename)
            except porcelain.PorcelainError as e:
                logger.exception("Error while handling %s: %r", filename, e)
                continue


class Back(WithRepoCommand):
    """Backport one or more files."""

    name = 'back'
    help = "Build and install one or more files."

    required_config_fields = ('target',)

    @classmethod
    def register_options(cls, parser):
        parser.add_argument('files', nargs='*', default=Default(tuple()),
            help="Backport selected files, all valid if empty.")
        super(Back, cls).register_options(parser)

    def run(self):
        p = porcelain.BackFile(self.env, self.active_repository)
        for filename in self._get_files(self.env.get('files')):
            try:
                p.handle(filename)
            except porcelain.PorcelainError as e:
                logger.exception("Error while handling %s: %r", filename, e)
                continue


class Diff(WithRepoCommand):
    """Check whether installed file are compatible with sources."""

    name = 'diff'
    help = "Compute diff between source and installed version of one or more files."

    required_config_fields = ('target',)

    @classmethod
    def register_options(cls, parser):
        parser.add_argument('files', nargs='*', default=Default(tuple()),
            help="Compute diff of selected files, all valid if empty.")
        super(Diff, cls).register_options(parser)

    def run(self):
        p = porcelain.DiffFile(self.env, self.active_repository)
        for filename in self._get_files(self.env.get('files')):
            try:
                p.handle(filename)
            except porcelain.PorcelainError as e:
                logger.exception("Error while handling %s: %r", filename, e)
                continue


class BackDiff(WithRepoCommand):
    """Check whether source file are compatible with installed version."""

    name = 'backdiff'
    help = "Compute diff between source and installed version of one or more files."

    required_config_fields = ('target',)

    @classmethod
    def register_options(cls, parser):
        parser.add_argument('files', nargs='*', default=Default(tuple()),
            help="Compute backward diff of selected files, all valid if empty.")
        super(BackDiff, cls).register_options(parser)

    def run(self):
        p = porcelain.BackDiffFile(self.env, self.active_repository)
        for filename in self._get_files(self.env.get('files')):
            try:
                p.handle(filename)
            except porcelain.PorcelainError as e:
                logger.exception("Error while handling %s: %r", filename, e)
                continue


class ImportFile(WithRepoCommand):
    name = 'import'
    help = "Import a new file into the repository"

    @classmethod
    def register_options(cls, parser):
        parser.add_argument('files', nargs='+', help="Add the selected files")
        parser.add_argument('--categories', required=True,
            help="Import into the selected category")
        parser.add_argument('--action', nargs='?',
            help="Build with the selected action")
        parser.add_argument('--action-params', nargs='*', default=Default(()),
            help="Extra parameters for the selected action")
        destination_group = parser.add_mutually_exclusive_group()
        destination_group.add_argument('--folder',
            help="Store the files in the given folder")
        super(ImportFile, cls).register_options(parser)

    def run(self):
        action_params = list(self.env.get('action_params', ()))

        p = porcelain.ImportFiles(self.env, self.active_repository)
        p.handle(
            files=self.env.get('files'),
            categories=self.env.get('categories'),
            action=self.env.get('action'),
            action_params=self.env.get('action_params', ()),
            folder=self.env.get('folder'),
        )


class RenameFile(WithRepoCommand):
    name = 'mv'
    help = "Rename a file within the repository"

    @classmethod
    def register_options(cls, parser):
        parser.add_argument('source', help="The file to rename")
        parser.add_argument('dest', help="The new path")
        super(RenameFile, cls).register_options(parser)

    def run(self):
        source = self.env.get('source')
        dest = self.env.get('dest')
        p = porcelain.RenameFile(self.env, self.active_repository)
        p.handle(source, dest)


class ListFiles(WithRepoCommand):
    name = 'files'
    help = "List all registered files"

    def run(self):
        target = self.env.target
        for filename in sorted(self.active_repository.iter_files()):
            file_config = self.active_repository.get_file_config(filename)
            self.info("%s -> %s", filename,
                file_config.get_destination(filename, target))


class ListCategories(WithRepoCommand):
    name = 'categories'
    help = "List all active categories"

    def run(self):
        for category in sorted(self.active_repository.categories):
            self.info(category)


base_commands = [
    HelpCommand,
    VersionCommand,
    Init,
    ListFiles,
    ListCategories,
    ImportFile,
    RenameFile,
    Make,
    Back,
    Diff,
    BackDiff,
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

import unittest

from uconf import converter


class LineTestCase(unittest.TestCase):
    def test_repr(self):
        self.assertEqual("Line('foo', 'bar')",
            repr(converter.Line('foo', 'bar')))

    def test_equality(self):
        self.assertEqual(
            converter.Line('foo', 'bar'),
            converter.Line('foo', 'bar'))

        self.assertNotEqual(
            converter.Line('foo', 'bar'),
            converter.Line('foo', 'baz'))

        self.assertNotEqual(
            converter.Line('foo', 'bar'),
            converter.Line('fo', 'bar'))

    def test_compare_to_other(self):
        self.assertNotEqual('foo', converter.Line('foo', 'bar'))
        self.assertNotEqual(converter.Line('foo', 'bar'), 'foo')

    def test_hash(self):
        s = set()
        for _i in range(5):
            s.add(converter.Line('foo', 'bar'))

        self.assertEqual(1, len(s))
        self.assertEqual(set([converter.Line('foo', 'bar')]), s)

    def test_fill_original_normal(self):
        l = converter.Line('foo', None)
        self.assertEqual(None, l.original)
        l.fill_original()
        self.assertEqual('foo', l.original)

    def test_fill_original_comment(self):
        l = converter.Line('#@foo', None)
        self.assertEqual(None, l.original)
        l.fill_original()
        self.assertEqual('#@@foo', l.original)

        l = converter.Line('"@foo', None)
        self.assertEqual(None, l.original)
        l.fill_original()
        self.assertEqual('"@@foo', l.original)

        l = converter.Line('!@foo', None)
        self.assertEqual(None, l.original)
        l.fill_original()
        self.assertEqual('!@@foo', l.original)


class BlockStackTestCase(unittest.TestCase):
    def test_bool_value(self):
        s = converter.BlockStack()
        self.assertFalse(s)

        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        self.assertTrue(s)

    def test_len(self):
        s = converter.BlockStack()
        self.assertEqual(0, len(s))

        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        self.assertEqual(1, len(s))

    def test_empty_repr(self):
        self.assertEqual("<BlockStack: []>", repr(converter.BlockStack()))

    def test_simple_repr(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        self.assertEqual("<BlockStack: [Block('if', 1, True, {})]>", repr(s))

    def test_empty_published(self):
        s = converter.BlockStack()
        self.assertTrue(s.published)

    def test_simple_published(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        self.assertTrue(s.published)

    def test_simple_unpublished(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_IF, published=False, start_line=1)
        self.assertFalse(s.published)

    def test_stack_published(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        s.enter(converter.Block.KIND_IF, published=True, start_line=2)
        s.enter(converter.Block.KIND_IF, published=True, start_line=3)
        self.assertTrue(s.published)

    def test_stack_unpublished(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        s.enter(converter.Block.KIND_IF, published=False, start_line=2)
        s.enter(converter.Block.KIND_IF, published=True, start_line=3)
        self.assertFalse(s.published)

    def test_stack_unpublished_leave(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        s.enter(converter.Block.KIND_IF, published=True, start_line=2)
        s.enter(converter.Block.KIND_IF, published=False, start_line=3)
        s.enter(converter.Block.KIND_IF, published=True, start_line=4)
        self.assertFalse(s.published)
        s.leave(converter.Block.KIND_IF)
        self.assertFalse(s.published)
        s.leave(converter.Block.KIND_IF)
        self.assertTrue(s.published)

    def test_empty_nolookup(self):
        s = converter.BlockStack()
        self.assertRaises(KeyError, s.get, 'foo')

    def test_simple_keyerror(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_IF, published=True, start_line=1)
        self.assertRaises(KeyError, s.get, 'foo')

    def test_simple_lookup(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_WITH, context={'bar': 'baz'}, start_line=1)
        self.assertRaises(KeyError, s.get, 'foo')
        self.assertEqual('baz', s.get('bar'))

    def test_stack_lookup(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_WITH, context={'bar': 1}, start_line=1)
        s.enter(converter.Block.KIND_WITH, context={'baz': 4}, start_line=2)
        s.enter(converter.Block.KIND_WITH, context={'bar': 3}, start_line=3)
        self.assertRaises(KeyError, s.get, 'foo')
        self.assertEqual(4, s.get('baz'))
        self.assertEqual(3, s.get('bar'))

    def test_stack_lookup_leave(self):
        s = converter.BlockStack()
        s.enter(converter.Block.KIND_WITH, context={'bar': 1}, start_line=1)
        s.enter(converter.Block.KIND_WITH, context={'baz': 4}, start_line=2)
        s.enter(converter.Block.KIND_WITH, context={'bar': 3}, start_line=3)
        self.assertRaises(KeyError, s.get, 'foo')
        self.assertEqual(4, s.get('baz'))
        self.assertEqual(3, s.get('bar'))
        s.leave(converter.Block.KIND_WITH)
        self.assertEqual(4, s.get('baz'))
        self.assertEqual(1, s.get('bar'))
        s.leave(converter.Block.KIND_WITH)
        self.assertRaises(KeyError, s.get, 'baz')
        self.assertEqual(1, s.get('bar'))


class GeneratorTestCase(unittest.TestCase):
    def test_no_special(self):
        txt = [
            'foo',
            'bar',
            'baz',
        ]

        g = converter.Generator(txt, categories=[], fs=None)
        expected = [converter.Line(s, s) for s in txt]
        out = list(g)
        self.assertEqual(expected, out)

    def test_nonmatching_if(self):
        txt = [
            'foo',
            '#@if blah',
            'bar',
            '#@endif',
            'baz',
        ]

        g = converter.Generator(txt, categories=['blih'], fs=None)
        expected = [
            converter.Line('foo', 'foo'),
            converter.Line(None, '#@if blah'),
            converter.Line(None, 'bar'),
            converter.Line(None, '#@endif'),
            converter.Line('baz', 'baz'),
        ]
        out = list(g)
        self.assertEqual(expected, out)

    def test_unclosed_nonmatching_if(self):
        txt = [
            'foo',
            '#@if blah',
            'bar',
        ]

        g = converter.Generator(txt, categories=['blih'], fs=None)
        expected = [
            converter.Line('foo', 'foo'),
            converter.Line(None, '#@if blah'),
            converter.Line(None, 'bar'),
        ]
        out = list(g)
        self.assertEqual(expected, out)

    def test_matching_if(self):
        txt = [
            'foo',
            '#@if blah',
            'bar',
            '#@endif',
            'baz',
        ]

        g = converter.Generator(txt, categories=['blah'], fs=None)
        expected = [
            converter.Line('foo', 'foo'),
            converter.Line(None, '#@if blah'),
            converter.Line('bar', 'bar'),
            converter.Line(None, '#@endif'),
            converter.Line('baz', 'baz'),
        ]
        out = list(g)
        self.assertEqual(expected, out)

    def test_unclosed_matching_if(self):
        txt = [
            'foo',
            '#@if blah',
            'bar',
        ]

        g = converter.Generator(txt, categories=['blah'], fs=None)
        expected = [
            converter.Line('foo', 'foo'),
            converter.Line(None, '#@if blah'),
            converter.Line('bar', 'bar'),
        ]
        out = list(g)
        self.assertEqual(expected, out)

    def test_inner_block(self):
        txt = [
            'foo',
            '#@if blah',
            'bar',
            '#@if blih',
            'barbar',
            '#@endif',
            'bazbaz',
            '#@endif',
            'baz',
        ]

        g = converter.Generator(txt, categories=['blih'], fs=None)
        expected = [
            converter.Line('foo', 'foo'),
            converter.Line(None, '#@if blah'),
            converter.Line(None, 'bar'),
            converter.Line(None, '#@if blih'),
            converter.Line(None, 'barbar'),
            converter.Line(None, '#@endif'),
            converter.Line(None, 'bazbaz'),
            converter.Line(None, '#@endif'),
            converter.Line('baz', 'baz'),
        ]
        out = list(g)
        self.assertEqual(expected, out)



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

from __future__ import absolute_import, unicode_literals

"""Common action code."""

import functools
import os.path

from . import converter
from . import fs


def catch_fs_exceptions(fun):
    @functools.wraps(fun)
    def decorated(self, *args, **kwargs):
        try:
            return fun(self, *args, **kwargs)
        except fs.FSError as e:
            print("Error while performing %s.%s(%s -> %s): %s" % (
                self.__class__.__name__, fun.__name__,
                self.source, self.destination,
                e))
            raise
    return decorated


class BaseAction(object):
    def __init__(self, source, destination, env, **kwargs):
        self.source = source
        self.destination = destination
        self.env = env
        self.fs = None

    @catch_fs_exceptions
    def forward(self, categories):
        """Apply the action."""
        self.fs = self.env.get_forward_fs()
        self._ensure_dir_exists(self.destination)
        self._forward(categories)

    def _forward(self, categories):
        raise NotImplementedError()

    @catch_fs_exceptions
    def backward(self, categories):
        """Revert the action."""
        self.fs = self.env.get_backward_fs()
        self._ensure_dir_exists(self.source)
        self._backward(categories)

    def _backward(self, categories):
        raise NotImplementedError()

    @catch_fs_exceptions
    def diff(self, categories):
        """Compute the differences between planned and actual file content."""
        self.fs = self.env.get_forward_fs()
        return self._diff(categories)

    def _diff(self, categories):
        raise NotImplementedError()

    @catch_fs_exceptions
    def backdiff(self, categories):
        """Compute the differences between original and backported file content."""
        self.fs = self.env.get_backward_fs()
        return self._backdiff(categories)

    def _backdiff(self, categories):
        raise NotImplementedError()

    def _ensure_dir_exists(self, path):
        dirname = os.path.dirname(path)
        self.fs.makedirs(dirname)

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.source)


class CopyAction(BaseAction):
    def _forward(self, categories):
        self.fs.copy(self.source, self.destination)

    def _backward(self, categories):
        self.fs.copy(self.destination, self.source)

    def _diff(self, categories):
        source_hash = self.fs.get_hash(self.source)
        dest_hash = self.fs.get_hash(self.destination)
        return [source_hash.hexdigest()], [dest_hash.hexdigest()]

    def _backdiff(self, categories):
        source, dest = self._diff(categories)
        # Simply reverse the diff
        return dest, source


class SymLinkAction(BaseAction):
    def _forward(self, categories):
        self.fs.create_symlink(
            link_name=self.destination,
            link_target=self.source,
            relative=False,
            force=False)


class FileContentAction(BaseAction):
    """An action based on file *contents*."""

    def _forward(self, categories):
        source_lines = self._readlines(self.source)
        destination_lines = self.forward_content(source_lines, categories)

        self.fs.writelines(self.destination, destination_lines)

    def forward_content(self, source_lines, categories):
        """Convert the source file, based on its lines.

        Args:
            source_lines (str list): lines of the original file
            categories (str list): active categories

        Yields:
            str: lines of the destination file
        """
        raise NotImplementedError()

    def _backward(self, categories):
        source_lines = self._readlines(self.source)
        modified_lines = self._readlines(self.destination)
        updated_lines = self.backward_content(source_lines, categories, modified_lines)

        self.fs.writelines(self.source, updated_lines)

    def backward_content(self, source_lines, categories, modified_lines):
        """Convert back the modified file, based on its lines.

        Args:
            source_lines (str list): lines of the original file
            categories (str list): active categories
            modified_lines (str list): lines of the modified files

        Yields:
            str: new lines for the source file
        """
        raise NotImplementedError()

    def _readlines(self, path, ignore_empty=True):
        if ignore_empty and not self.fs.file_exists(path):
            return []
        return self.fs.readlines(path)

    def _diff(self, categories):
        source_lines = self._readlines(self.source)
        planned_lines = self.forward_content(source_lines, categories)
        actual_lines = self._readlines(self.destination)

        return list(planned_lines), list(actual_lines)

    def _backdiff(self, categories):
        source_lines = list(self._readlines(self.source))
        destination_lines = list(self._readlines(self.destination))
        backported_lines = self.backward_content(
                source_lines, categories, destination_lines)

        return list(backported_lines), source_lines


class FileProcessingAction(FileContentAction):
    """Process a file, using usual rules."""
    def forward_content(self, source_lines, categories):
        processor = converter.FileProcessor(source_lines, self.fs)
        return processor.forward(categories)

    def backward_content(self, source_lines, categories, modified_lines):
        processor = converter.FileProcessor(source_lines, self.fs)
        return processor.backward(categories, modified_lines)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

from __future__ import unicode_literals, absolute_import

import os
import socket


def filter_iter(iterator, items, key=lambda o: o, empty_is_all=False):
    """Filter items from an iterator, keeping only those in a set."""

    output_all = False
    if items:
        items = frozenset(items)
    elif empty_is_all:
        output_all = True

    for item in iterator:
        if output_all or key(item) in items:
            yield item


def flatten(fields, separator=None):
    """Flatten a list of space-separated names."""
    flattened = set()
    for field in fields:
        flattened |= set(field.split(separator))
    return flattened


def get_absolute_path(path, base=''):
    path = os.path.join(base, os.path.expanduser(path))
    return os.path.abspath(path)


def get_relative_path(root, path, base=''):
    """Converts the given 'path' to a path relative to the 'root'."""
    return os.path.relpath(get_absolute_path(path, base=base), root)


def get_hostnames(name=''):
    """Return the list of hostnames for a name (defaults to local host)."""
    fqdn = socket.getfqdn(name)
    return fqdn, fqdn.split('.')[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

import re
import unittest

from uconf import action_parser
from uconf import rule_parser

class RuleLexerTestCase(unittest.TestCase):

    def setUp(self):
        self.rule_lexer = rule_parser.RuleLexer()

    def test_simple(self):
        rules = (
            ('a', rule_parser._TextNode('a')),
            ('a && b', rule_parser._AndNode(
                [rule_parser._TextNode('a'), rule_parser._TextNode('b')])),
            ('a || b', rule_parser._OrNode(
                [rule_parser._TextNode('a'), rule_parser._TextNode('b')])),
            ('!a', rule_parser._NegateNode(rule_parser._TextNode('a'))),
            ('a || (b || c)', rule_parser._OrNode([
                rule_parser._TextNode('a'),
                rule_parser._TextNode('b'),
                rule_parser._TextNode('c'),
                ])),
            ('a && !b', rule_parser._AndNode([
                rule_parser._TextNode('a'),
                rule_parser._NegateNode(rule_parser._TextNode('b')),
                ])),
        )

        for rule_text, expected_node in rules:
            rule = self.rule_lexer.get_rule(rule_text)
            self.assertEqual(expected_node, rule.node)


class ActionLexerTestCase(unittest.TestCase):
    def setUp(self):
        self.action_lexer = action_parser.ActionLexer()

    def test_single_quoted_regexps(self):
        samples = (
            (r"'foo'", r"'foo'"),
            (r"'foo\"bar'", r"'foo\"bar'"),
            (r"'foo\bar'", r"'foo\bar'"),
            (r"'foo\'bar'", r"'foo\'bar'"),
            (r"'foo'bar'", r"'foo'"),
            (r"'foo\\'bar'", r"'foo\\'"),
        )

        regexp = re.compile(action_parser.SingleQuotedTextToken.regexp)

        for source, target in samples:
            m = regexp.match(source)
            self.assertIsNotNone(m, "No match found in %r" % source)
            self.assertEqual(m.group(), target)

    def test_double_quoted_regexps(self):
        samples = (
            (r'"foo"', r'"foo"'),
            (r'"foo\'bar"', r'"foo\'bar"'),
            (r'"foo\bar"', r'"foo\bar"'),
            (r'"foo\"bar"', r'"foo\"bar"'),
            (r'"foo"bar"', r'"foo"'),
            (r'"foo\\"bar"', r'"foo\\"'),
        )

        regexp = re.compile(action_parser.DoubleQuotedTextToken.regexp)

        for source, target in samples:
            m = regexp.match(source)
            self.assertIsNotNone(m, "No match found in %r" % source)
            self.assertEqual(m.group(), target)

    def test_simple(self):
        samples = (
            (r'foo', {'foo': None}),
            (r'foo bar', {'foo': None, 'bar': None}),
            (r' foo', {'foo': None}),
            (r'  foo   ', {'foo': None}),
            (r'foo   bar', {'foo': None, 'bar': None}),
            (r'foo=bar bar', {'foo': 'bar', 'bar': None}),
            (r'foo=bar bar=foo foo=bar', {'foo': 'bar', 'bar': 'foo'}),
            (r'foo="bar"', {'foo': 'bar'}),
            (r'foo=" bar "', {'foo': ' bar '}),
            (r"foo='bar'", {'foo': 'bar'}),
            (r"foo=' bar '", {'foo': ' bar '}),
            (r'foo="bar\"baz"', {'foo': 'bar"baz'}),
            (r"""foo="bar\"baz" bar='baz\'foo'""",
                {'foo': 'bar"baz', 'bar': 'baz\'foo'}),
        )

        for text, expected in samples:
            options = self.action_lexer.get_options(text)
            self.assertEqual(options, expected)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.


from __future__ import absolute_import, unicode_literals


"""Low level actions for uconf."""


import difflib
import logging
import os.path

from . import helpers

class PorcelainError(Exception):
    def __init__(self, user_message):
        self.user_message = user_message
        super(PorcelainError, self).__init__()


class Porcelain(object):
    def __init__(self, env, active_repo=None):
        self.env = env
        self.active_repo = active_repo
        self.logger = logging.getLogger(
            '%s.%s' % (__name__, self.__class__.__name__))

    def handle(self, *args, **kwargs):
        """Run the porcelain command.

        Raises:
            PorcelainError: if anything went wrong.
        """
        raise NotImplementedError()


class ImportFiles(Porcelain):
    def _make_action_text(self, action=None, action_params=()):
        if action:
            text = action
        elif action_params:  # Default action, but overridden params
            text = 'parse'
        else:
            return ''

        if action_params:
            text = '%s %s' % (text,
                ' '.join('%s=%s' % param for param in action_params))
        return text

    def _prepare_files(self, targets, storage_folder=None):
        """Converts a set of files to a list of (storage, target) pairs.

        The returned (storage, target) pairs are relative paths:
            - the 'storage' path is relative to the repository root
            - the 'target' path is relative to the install root
        """
        target_root = self.env.target

        targets = [
            helpers.get_relative_path(target_root, target, base=target_root)
            for target in targets]

        if storage_folder:
            storage_folder = helpers.get_relative_path(
                self.env.root, storage_folder, base=self.env.root)

            return [
                (os.path.join(storage_folder, os.path.basename(t)), t)
                for t in targets]

        else:
            return [(t, t) for t in targets]

    def handle(self, files, categories, action=None, action_params=(),
            folder=None, *args, **kwargs):
        files_config = self.env.repository.files_config
        actions_config = self.env.repository.actions_config

        # Cleanup paths
        paths = self._prepare_files(files, storage_folder=folder)
        files_text = ' '.join(storage for storage, _install in paths)

        # Add to the 'files' section
        if files_text in list(files_config[categories]):
            self.logger.warning("Files '%s' already registered for categories %r",
                files_text, categories)
        else:
            self.logger.info("Registering files '%s' for categories %r",
                files_text, categories)
            files_config.add(categories, files_text)

        # Handle dedicated action/options for files
        for storage, install in paths:
            file_params = list(action_params)
            if storage != install:
                file_params.append(('dest', install))

            action_text = self._make_action_text(action, file_params)
            if action_text:
                self.logger.info("Adding rule %r for files '%s'",
                    action_text, storage)
                actions_config[storage] = action_text

        # Write out the configuration
        self.env.repository.write_config(self.env.get_uconf_fs())


class RenameFile(Porcelain):
    """Rename a file within the repository."""

    def _rename_in_files(self, source, dest):
        repo_files = self.env.repository.files_config
        updates = []
        for categories, file_lists in repo_files.items():
            if source in helpers.flatten(file_lists):
                updates.append((categories, file_lists))

        for categories, old_files in updates:
            new_files = [fl.replace(source, dest) for fl in old_files]
            repo_files[categories] = new_files

    def _rename_in_actions(self, source, dest):
        """Rename the file in the 'actions' section of the repo's config."""
        repo_actions = self.env.repository.actions_config
        if source in repo_actions:
            repo_actions[dest] = repo_actions.pop(source)

    def handle(self, source, dest, *args, **kwargs):
        """Renames a file within the repository, both on disk and in config.

        The source and dest arguments should be relative to the repository root.
        """
        source = helpers.get_relative_path(self.env.root, source, base=self.env.root)
        dest = helpers.get_relative_path(self.env.root, dest, base=self.env.root)

        self._rename_in_files(source, dest)
        self._rename_in_actions(source, dest)

        abs_source = helpers.get_absolute_path(source, base=self.env.root)
        abs_dest = helpers.get_absolute_path(dest, base=self.env.root)

        fs = self.env.get_repo_fs()
        fs.rename(abs_source, abs_dest)
        self.env.repository.write_config(fs)


class FilePorcelain(Porcelain):
    """Porcelain command for a single file."""

    def handle(self, filename, *args, **kwargs):
        if self.active_repo is None:
            raise PorcelainError("This porcelain command requires an active repository.")

        try:
            file_config = self.active_repo.get_file_config(filename,
                    default_action=self.env.get('default_action', 'parse'))
        except KeyError:
            raise PorcelainError("File %s not in repository." % filename)

        self.handle_file(filename, file_config, *args, **kwargs)


class MakeFile(FilePorcelain):
    def handle_file(self, filename, file_config):
        action = file_config.get_action(filename, self.env)
        self.logger.info("Building file %s (%s)", filename, action.__class__.__name__)
        action.forward(self.active_repo.categories)


class BackFile(FilePorcelain):
    def handle_file(self, filename, file_config):
        action = file_config.get_action(filename, self.env)
        self.logger.info("Backporting file %s (%s)", filename, action.__class__.__name__)
        action.backward(self.active_repo.categories)


class DiffFile(FilePorcelain):
    def handle_file(self, filename, file_config):
        action = file_config.get_action(filename, self.env)
        planned, actual = action.diff(self.active_repo.categories)
        if planned != actual:
            diff = difflib.unified_diff(actual, planned,
                fromfile=action.destination, tofile=action.destination, lineterm='')
            diff = ('',) + tuple(diff)
            diff = '\n'.join(diff)
            self.logger.info("File %s has changed: %s", filename, diff)


class BackDiffFile(FilePorcelain):
    def handle_file(self, filename, file_config):
        action = file_config.get_action(filename, self.env)
        planned, actual = action.backdiff(self.active_repo.categories)
        if planned != actual:
            diff = difflib.unified_diff(actual, planned,
                fromfile=action.source, tofile=action.source, lineterm='')
            diff = ('',) + tuple(diff)
            diff = '\n'.join(diff)
            self.logger.info("File %s has changed: %s", filename, diff)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

from __future__ import unicode_literals, absolute_import

"""Abstract the filesystem layer."""

import hashlib
import io
import logging
import os
import stat

import fslib
import fslib.builders
import fslib.stacking


logger = logging.getLogger(__name__)


FSError = fslib.FSError


class FSLoader(object):
    def __init__(self, *write_paths, **kwargs):
        self.dry_run = kwargs.pop('dry_run', False)
        self.default_encoding = kwargs.pop('default_encoding', 'utf-8')
        self.write_paths = write_paths
        self.fs, self.subfs = self._prepare_fs(write_paths, dry_run=self.dry_run)

    def _prepare_fs(self, paths, dry_run=False):
        """Prepare the filesystem for a set of writable paths."""
        base_fs = fslib.stacking.MountFS()
        base = fslib.stacking.ReadOnlyFS(fslib.OSFS())
        base_fs.mount_fs(base, fslib.ROOT)

        sub_filesystems = {}

        for path in paths:
            subfs = fslib.OSFS(mapped_root=path)
            if dry_run:
                ro_fs = fslib.stacking.ReadOnlyFS(subfs)
                mem_fs = fslib.builders.make_memory_fake()
                union_fs = fslib.stacking.UnionFS()
                union_fs.add_branch(mem_fs, 'mem', rank=0, writable=True)
                union_fs.add_branch(ro_fs, 'os_ro', rank=1, writable=False)
                subfs = union_fs
                sub_filesystems[path] = mem_fs
            base_fs.mount_fs(subfs, path)

        return fslib.FileSystem(base_fs), sub_filesystems

    def get_changes(self):
        if self.dry_run:
            for path, fs in self.subfs.items():
                for subpath in fs.walkfiles():
                    length = len(fs.getcontents(subpath))
                    yield '%s%s' % (path, subpath), length

    def __del__(self):
        try:
            if self.dry_run:
                for path, lenth in self.get_changes():
                    logger.info("[Dry-run] Updated path %s (%d bytes)", path, lenth)
        except:
            # Something went wrong while deleting ourselves.
            pass

    def __getattr__(self, name):
        return getattr(self.fs, name)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.


from __future__ import unicode_literals


"""Parse command-line arguments.

The CLI parsing is a three-step process:
    1) Parse a minimal set of configuration options:
        - verbosity
        - config file
        - repository root
    2) If a config file or a repository root is provided, read it and
       merge its "[core]" section into defaults for the next step
    3) Perform the full parsing
"""


import argparse
import logging
import os
import stat

import confutils

from . import commands
from . import config
from . import constants
from . import helpers
from . import __version__


Default = confutils.Default


class CLI(object):
    """Command-line interface.

    Attributes:
        progname (str): name to use to refer to the program
        parser (argparse.ArgumentParser): list of available CLI args & options
        subparsers (argparse.SubParser): handles action-specific subparsers
    """

    def __init__(self, progname):
        self.progname = progname
        self.parser = self.make_base_parser(self.progname)
        self.subparsers = self.parser.add_subparsers(help="Commands", dest='subcommand')

        self.register_base_commands()

    # Options and parsers
    # -------------------

    def make_base_parser(self, progname):
        parser = argparse.ArgumentParser(prog=self.progname,
            argument_default=Default(None))
        parser.add_argument('--version', '-V', help="Display version", action='version',
            version='%(prog)s ' + __version__)
        return parser

    def register_options(self, parser):
        """Register global options"""
        parser.add_argument('--root', '-r', help="Set uconf repository root")
        parser.add_argument('--config-dir', '-c', help="Use uconf config dir CONFIG_DIR")
        parser.add_argument('--prefs', '-p', nargs='*', default=constants.CONFIG_FILES,
            help="Read user preferences from PREF files", metavar='PREF')
        parser.add_argument('--version', '-V', help="Display version", action='version',
            version='%(prog)s ' + __version__)
        parser.add_argument('--dry-run', '-n', help="Pretend to run the actions",
            action="store_true", default=Default(False))
        parser.add_argument('--initial', '-i', nargs='*',
            help="Set alternate initial categories", default=Default(tuple()))
        parser.add_argument('--strict', help="Stop on warnings",
            action="store_true", default=Default(False))
        parser.add_argument('--target', '-t', help="Write generated files to TARGET",
            default=Default(''))

    # Registering commands
    # --------------------

    def register_command(self, command_class):
        """Register a new command from its class."""
        cmd_parser = self.subparsers.add_parser(command_class.get_name(),
            help=command_class.get_help())
        self.register_options(cmd_parser)
        command_class.register_options(cmd_parser)
        cmd_parser.set_defaults(command=command_class)

    def register_base_commands(self):
        """Register all known, base commands."""
        for command_class in commands.base_commands:
            self.register_command(command_class)

    # Reading configuration
    # ----------------------

    def make_command_config(self, args, command_class):
        """Prepare the (merged) options pseudo-dict for a given command.

        Uses, in turn:
            - command-specific command line options
            - global command line options
            - command-specific configuration file options
            - global configuration file options
        """
        return config.Env.from_files(
            repo_root=args.root or os.getcwd(),
            sections=(command_class.get_name(),),
            extra=confutils.DictNamespace(args),
        )

    # Logging
    # -------

    def setup_logging(self):
        """Set up a minimal logging configuration."""
        root_logger = logging.getLogger()
        handler = logging.StreamHandler()
        root_logger.addHandler(handler)
        root_logger.setLevel(logging.INFO)

    # Running commands
    # ----------------

    def run_from_argv(self, argv):
        """Actually run the requested command from the argv."""
        self.setup_logging()
        # Add command-specific arguments
        args = self.parser.parse_args(argv)
        command_name = args.subcommand
        if command_name is None:
            self.parser.print_help()
            return
        command_class = args.command

        # Merge all pref bits
        env = self.make_command_config(args, command_class)

        # Build and run the command
        cmd = command_class(env, self.parser)
        return cmd.run()


def main(argv):
    """Run the prgoram."""
    progname = argv[0]
    args = argv[1:]
    cli = CLI(progname)
    return cli.run_from_argv(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

import codecs
import os
import re
import sys

from setuptools import setup

root_dir = os.path.abspath(os.path.dirname(__file__))


def get_version(package_name):
    version_re = re.compile(r"^__version__ = [\"']([\w_.-]+)[\"']$")
    package_components = package_name.split('.')
    init_path = os.path.join(root_dir, *(package_components + ['__init__.py']))
    with codecs.open(init_path, 'r', 'utf-8') as f:
        for line in f:
            match = version_re.match(line[:-1])
            if match:
                return match.groups()[0]
    return '0.1.0'


PACKAGE = 'uconf'


setup(
    name=PACKAGE,
    version=get_version(PACKAGE),
    description="UConf, a smart tool for managing config files",
    long_description=''.join(codecs.open('README.rst', 'r', 'utf-8').readlines()),
    author="Raphaël Barrois",
    author_email="raphael.barrois+%s@polytechnique.org" % PACKAGE,
    license="BSD",
    keywords=['configuration', 'management', 'uconf', 'confmgr', 'config'],
    url="https://github.com/rbarrois/%s/" % PACKAGE,
    download_url="https://pypi.python.org/pypi/%s/" % PACKAGE,
    packages=[PACKAGE],
    platforms=["OS Independent"],
    scripts=['bin/uconf'],
    install_requires=codecs.open('requirements.txt', 'r', 'utf-8').readlines(),
    setup_requires=[
        'setuptools>=0.8',
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Systems Administration",
    ],
    test_suite='tests',
)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.


from __future__ import unicode_literals, absolute_import


"""config.h-style file holding all global constants for uconf."""


CONFIG_FILES = ('/etc/uconf.conf', '~/.uconfrc')
REPO_SUBFOLDER = '.uconf'



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

from __future__ import unicode_literals, absolute_import

"""Py2 backwards compatibility fixes."""

import sys

PY2 = sys.version_info[0] < 3

if PY2:
    text_types = (str, unicode)
else:
    text_types = (str,)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

import difflib
import re

from uconf import rule_parser


class FileProcessor(object):
    """Handles 'standard' processing of a file.

    Attributes:
        src (str list): lines of the file to process
        fs (FileSystem): abstraction toward the filesystem
    """
    def __init__(self, src, fs):
        self.src = list(src)
        self.fs = fs

    def _get_gen_config(self, categories):
        return GeneratorConfig(
            categories=categories,
            commands=[cmd() for cmd in DEFAULT_COMMANDS],
            fs=self.fs,
        )

    def forward(self, categories):
        """Process the source file with an active list of categories."""
        gen_config = self._get_gen_config(categories)
        generator = gen_config.load(self.src)
        for line in generator:
            if line.output is not None:
                yield line.output

    def backward(self, categories, modified):
        """Revert a file.

        Args:
            categories (str iterable): active categories
            modified (str list): lines of the modified file

        Yields:
            str: updated lines for the original file
        """
        categories = frozenset(categories)
        original_output = self.forward(categories)
        diff = Differ(original_output, modified)
        gen_config = self._get_gen_config(categories)
        generator = gen_config.load(self.src)
        backporter = Backporter(diff, generator)

        for line in backporter:
            yield line


class Differ(object):
    """Computes differences between two files (as string lists).

    Based on difflib.SequenceMatcher, but yields atomic operations.

    Attributes:
        original (str list): lines of the original file
        modified (str list): lines of the modified file
    """
    def __init__(self, original, modified):
        self.original = list(original)
        self.modified = list(modified)

    def __iter__(self):
        """Yield atomic diff lines.

        Yields:
            (operation, new_line) tuples.
        """
        matcher = difflib.SequenceMatcher(a=self.original, b=self.modified)
        opcodes = matcher.get_opcodes()
        for opcode, original_i, original_j, modified_i, modified_j in opcodes:
            if opcode == 'equal':
                for original_lineno in range(original_i, original_j):
                    yield (opcode, self.original[original_lineno])
            elif opcode == 'insert':
                for modified_lineno in range(modified_i, modified_j):
                    yield (opcode, self.modified[modified_lineno])
            elif opcode == 'delete':
                for original_lineno in range(original_i, original_j):
                    yield (opcode, self.original[original_lineno])
            elif opcode == 'replace':
                common = min(original_j - original_i, modified_j - modified_i)
                for modified_lineno in range(modified_i, modified_i + common):
                    yield (opcode, self.modified[modified_lineno])
                for modified_lineno in range(modified_i + common, modified_j):
                    yield ('insert', self.modified[modified_lineno])
                for original_lineno in range(original_i + common, original_j):
                    yield ('delete', self.original[original_lineno])


class Backporter(object):
    """Handles the backporting of a diff to an original file.

    Attributes:
        diff ((operation, new_line) iterable): the lines of the diff
        source (Line iterable): the lines of the source
    """

    def __init__(self, diff, source):
        self.diff = diff
        self.source = source

    def reverse(self, output):
        """Convert back an output line into its original version."""
        line = Line(output, None)
        line.fill_original()
        return line.original

    def __iter__(self):
        """Yield lines from the initial file."""
        diff = iter(self.diff)

        for line in self.source:
            # Loop through the generated lines

            if line.output is None:
                # Masked line (comment, command)
                # Always include
                yield line.original

            else:
                action, output = diff.next()
                while action == 'insert':
                    # Inserting lines
                    # Always include, without forwarding the source
                    yield self.reverse(output)
                    action, output = diff.next()

                if action == 'delete':
                    # Deleting one line, advance the source
                    continue
                else:
                    # Replacing / equal, write the current line
                    yield self.reverse(output)

        # Handle additional lines from the diff
        # Should only be 'insert' lines.
        for action, output in diff:
            assert action == 'insert', "Unexpected action %s on %r" % (action, output)
            yield self.reverse(output)


class Line(object):
    def __init__(self, output, original):
        self.output = output
        self.original = original

    def __repr__(self):
        return "Line(%r, %r)" % (self.output, self.original)

    def __hash__(self):
        return hash((self.output, self.original))

    def __eq__(self, other):
        if not isinstance(other, Line):
            return NotImplemented
        return self.output == other.output and self.original == other.original

    def fill_original(self):
        """Fill the 'original' part from the output."""
        if self.original is not None:
            return

        # If the output line looks like a comment or command, escape it.
        should_escape_re = re.compile(r'^["!#]@')
        if should_escape_re.match(self.output):
            self.original = '%s@%s' % (
                self.output[:2],
                self.output[2:],
            )
        else:
            self.original = self.output


class Block(object):
    KIND_IF = 'if'
    KIND_WITH = 'with'

    def __init__(self, kind, start_line, published=True, context=None):
        self.kind = kind
        self.published = published
        self.context = context or {}
        self.start_line = start_line

    def __repr__(self):
        return "Block(%r, %d, %r, %r)" % (
            self.kind,
            self.start_line,
            self.published,
            self.context,
        )


class BlockStack(object):
    def __init__(self):
        self.blocks = []

    def __nonzero__(self):
        return bool(self.blocks)

    def __len__(self):
        return len(self.blocks)

    def __repr__(self):
        return "<BlockStack: %r>" % self.blocks

    @property
    def published(self):
        return all(b.published for b in self.blocks)

    @property
    def merged_context(self):
        context = {}
        for block in self.blocks:
            context.update(block.context)
        return context

    def enter(self, *args, **kwargs):
        block = Block(*args, **kwargs)
        self.blocks.append(block)
        return block

    def leave(self, kind):
        if not self.blocks:
            raise ValueError("Not inside a block.")
        last_kind = self.blocks[-1].kind
        if last_kind != kind:
            raise ValueError("Unexpected last block kind: %s!=%s." %
                (last_kind, kind))
        return self.blocks.pop()


class BaseCommand(object):
    """A command.

    Entry points: get_keys(), handle(...).
    """
    keys = ()

    def get_keys(self):
        """Return the list of "keys" (or "commands") handled by this class."""
        return self.keys

    def handle(self, key, argline, state, config):
        """Handle a line.

        Args:
            key (str): one of the keys in get_keys()
            argline (str): everything after the key and a space
            state (GeneratorState): the current state of the generator
            config (GeneratorConfig): various config-time params of the generator
        """
        raise NotImplementedError()


class BaseBlockCommand(BaseCommand):
    enter_keys = ()
    inside_keys = ()
    exit_keys = ()

    def get_keys(self):
        return self.enter_keys + self.inside_keys + self.exit_keys

    def handle(self, key, argline, state, config):
        if key in self.enter_keys:
            return self.enter(key, argline, state, config)
        elif key in self.inside_keys:
            return self.inside(key, argline, state, config)
        else:
            assert key in self.exit_keys
            return self.exit(key, argline, state, config)

    def enter(self, key, argline, state, config):
        raise NotImplementedError()

    def inside(self, key, argline, state, config):
        raise NotImplementedError()

    def exit(self, key, argline, state, config):
        raise NotImplementedError()


class IfBlockCommand(BaseBlockCommand):
    enter_keys = ('if',)
    inside_keys = ('else', 'elif')
    exit_keys = ('endif',)

    def __init__(self, **kwargs):
        super(IfBlockCommand, self).__init__(**kwargs)
        self.rule_lexer = rule_parser.RuleLexer()

    def enter(self, key, argline, state, config):
        rule = self.rule_lexer.get_rule(argline)
        state.enter_block(Block.KIND_IF, published=rule.test(config.categories))

    def inside(self, key, argline, state, config):
        if key == 'else':
            if argline:
                state.error("Command 'else' takes no argument, got %r", argline)

            last_block = state.leave_block(Block.KIND_IF)
            state.enter_block(Block.KIND_IF, published=not last_block.published)
        else:
            assert key == 'elif'
            last_block = state.leave_block(Block.KIND_IF)
            if last_block.published:
                published = False
            else:
                rule = self.rule_lexer.get_rule(argline)
                published = rule.test(config.categories)

            state.enter_block(Block.KIND_IF, published=published)

    def exit(self, key, argline, state, config):
        assert key == 'endif'
        if argline:
            state.error("Command 'endif' takes no argument, got %r", argline)
        state.leave_block(Block.KIND_IF)


class WithBlockCommand(BaseBlockCommand):
    enter_keys = ('with', 'withfile')
    exit_keys = ('endwith',)

    with_args_re = re.compile(r'^(\w+)=(.*)$')

    def _read_file(self, filename, config):
        """Read one line from a file."""
        return config.fs.read_one_line(filename)

    def _parse_with_args(self, args, state, config):
        """Parce "#@with" arguments (and validate the line structure)."""
        match = self.with_args_re.match(args)
        if not match:
            state.error("Invalid 'with' argument %r", args)
        return match.groups()

    def enter(self, key, argline, state, config):
        if key == 'with':
            var, value = self._parse_with_args(argline, state=state)
            state.enter_block(Block.KIND_WITH, context={var: value})
        else:
            assert key == 'withfile'
            var, filename = self._parse_with_args(argline, state=state)
            value = self._read_file(filename, config)
            state.enter_block(Block.KIND_WITH, context={var: value})

    def exit(self, key, argline, state, config):
        last_block = state.leave_block(Block.KIND_WITH)
        if argline and argline not in last_block.context:
            raise CommandError("Block mismatch: closing 'with' block from line %d with invalid variable %r" % (last_block.start_line, argline))
        del state.context[var]


class GeneratorState(object):
    """Handles the internal generator state.

    Attributes:
        in_published_block (bool): whether the current block should be published
        context (str => str dict): maps a placeholder name to its content
        _current_lineno (int): the current line number
    """

    def __init__(self):
        self.block_stack = BlockStack()

        self._current_lineno = 0

    @property
    def in_published_block(self):
        return self.block_stack.published

    @property
    def context(self):
        return self.block_stack.merged_context

    def error(self, message, *args):
        err_msg = "Error on line %d: " % self._current_lineno
        raise ValueError(error + message % args)

    def advance_to(self, lineno):
        self._current_lineno = lineno

    def enter_block(self, kind, published=True, context=None):
        return self.block_stack.enter(
            kind=kind,
            published=published,
            context=context,
            start_line=self._current_lineno,
        )

    def leave_block(self, kind):
        try:
            return self.block_stack.leave(kind)
        except ValueError as e:
            self.invalid("Error when closing block: %r", e)



DEFAULT_COMMANDS = [
    IfBlockCommand,
    WithBlockCommand,
]


class Generator(object):
    """Generate the output from a source.

    Attributes:
        src (iterable of str): the source lines
        state (GeneratorState): the current generator state
    """

    command_prefix_re = re.compile(r'^(["!#]@)(.+)$')

    def __init__(self, src, commands, config):
        self.src = src
        self.config = config
        self.state = GeneratorState()
        self.commands_by_key = {}
        for command in commands:
            for key in command.get_keys():
                if key in self.commands_by_key:
                    raise ValueError("Duplicate command for key %s: got %r and %r"
                        % (key, command, self.commands_by_key[key]))
                self.commands_by_key[key] = command

    def __iter__(self):
        for lineno, line in enumerate(self.src):
            self.state.advance_to(lineno)

            match = self.command_prefix_re.match(line)
            if match:
                prefix, command = match.groups()
                output = self.handle_line(prefix, command)

            # If displaying the line, replace placeholders.
            elif self.state.in_published_block:
                updated_line = line
                for var, value in self.state.context.items():
                    pattern = '@@%s@@' % var
                    updated_line = updated_line.replace(pattern, value)
                output = updated_line

            # Not displaying the line
            else:
                output = None

            yield Line(output, line)


    def handle_line(self, prefix, command):
        if command.startswith('#'):
            # A comment
            return None

        elif command.startswith('@'):
            # An escaped line
            return prefix + line[1:]

        else:
            if ' ' in command:
                name, args = command.split(' ', 1)
            else:
                name, args = command, ''
            self.handle_command(name, args)
            return None

    def handle_command(self, command, args):
        """Handle a "#@<command>" line."""
        if command not in self.commands_by_key:
            raise CommandError("Unknown command '%s' (not in %r)" % (command, sorted(self.commands_by_key)))

        handler = self.commands_by_key[command]
        handler.handle(command, args, self.state, self.config)


class GeneratorConfig(object):
    def __init__(self, categories, commands, fs, generator=Generator):
        self.categories = categories
        self.commands = commands
        self.fs = fs
        self.fs_root = '/'
        self.generator_class = generator

    def load(self, source_file):
        return self.generator_class(source_file,
            config=self,
            commands=self.commands,
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.

from __future__ import unicode_literals, absolute_import

"""Handle repository-wide configuration.

These settings only relate to the list of categories, files, ... registered in
the repository, regardless of current settings.
"""


import fnmatch
import os
import stat
import time

import confutils

from . import action_parser
from . import actions
from . import compat
from . import constants
from . import fs
from . import helpers
from . import rule_parser


class FileConfig(object):
    """Definition of the action for a file."""

    COPY = 'copy'
    SYMLINK = 'symlink'
    PARSE = 'parse'

    ACTIONS = {
        COPY: actions.CopyAction,
        SYMLINK: actions.SymLinkAction,
        PARSE: actions.FileProcessingAction,
    }

    def __init__(self, action, **options):
        if action not in self.ACTIONS:
            raise ValueError("Invalid action %s, choose one of %s" % (
                action, ', '.join(self.ACTIONS)))
        self.action = action
        self.options = options

    def get_destination(self, filename, target=''):
        """Find the appropriate target for a file."""
        if 'dest' in self.options:
            # Explicit target path
            destination = self.options['dest']
        elif 'destdir' in self.options:
            # Replaced destination folder
            destination = os.path.join(self.options['destdir'],
                os.path.basename(filename))
        else:
            destination = filename

        return helpers.get_absolute_path(destination, base=target)

    def get_action(self, filename, env):
        action = self.ACTIONS[self.action]
        abs_source = helpers.get_absolute_path(filename, base=env.root)
        abs_dest = self.get_destination(filename, env.target)

        return action(source=abs_source, destination=abs_dest,
            env=env, **self.options)

    def __repr__(self):
        return '<FileConfig: %s %r>' % (self.action, self.options)


class GlobStore(object):
    def __init__(self, *args):
        self.entries = list(*args)

    def __getitem__(self, key):
        for glob, value in self.entries:
            if fnmatch.fnmatchcase(key, glob):
                return value
        raise KeyError("Key %s not found in %r" % (key, self))

    def __setitem__(self, key, value):
        for i, entry in enumerate(self.entries):
            if key == entry[0]:
                self.entries[i] = (key, value)
        self.entries.append((key, value))

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __repr__(self):
        return "GlobStore(%r)" % self.entries


class RepositoryView(object):
    """A repository, as seen for a given set of initial categories.

    Attributes:
        base: the Repository on which this view is based
        categories: frozenset of active category names
    """

    def __init__(self, base):
        self.base = base
        self.categories = frozenset()

    def set_initial_categories(self, initial):
        self.categories = frozenset(initial)
        for category_rule, extra_categories in self.base.category_rules:
            if category_rule.test(self.categories):
                self.categories |= extra_categories

    def iter_files(self):
        """Retrieve all active files for this view

        Yields:
            filename
        """
        for file_rule, filename in self.base.file_rules:
            if file_rule.test(self.categories):
                yield filename

    def get_file_config(self, filename, default_action='parse'):
        default_config = FileConfig(default_action)
        return self.base.file_configs.get(filename, default_config)


class Repository(object):
    """Holds repository configuration.

    Attributes:
        categories (str set): all categories
        files (str list): all managed files
        file_actions (dict(str => FileConfig)): actions for files
        rule_lexer (rule_parser.RuleLexer): lexer to use for rule parsing
    """

    def __init__(self, root=None, *args, **kwargs):
        self.root = root
        self.config = confutils.ConfigFile()
        self.actions_config = self.config.section_view('actions')
        self.files_config = self.config.section_view('files', True)
        self.categories_config = self.config.section_view('categories', True)

        self.category_rules = []
        self.file_rules = []
        self.file_configs = GlobStore()
        self.rule_lexer = rule_parser.RuleLexer()
        self.action_lexer = action_parser.ActionLexer()

        self._read_config()

    @property
    def uconf_dir(self):
        return os.path.join(self.root, constants.REPO_SUBFOLDER)

    @property
    def config_path(self):
        return os.path.join(self.uconf_dir, 'config')

    def extract(self, initial):
        """Extract a 'view' on this repository for given initial categories."""
        view = RepositoryView(self)
        view.set_initial_categories(initial)
        return view

    def write_config(self, fs):
        """Update the configuration."""
        temp_name = '.config-%s.new' % time.strftime('%Y%m%d%H%M%S')
        temp_path = os.path.join(self.uconf_dir, temp_name)

        with fs.open(temp_path, 'wt') as f:
            self.config.write(f)

        fs.rename(temp_path, self.config_path)

    def _read_config(self):
        if not self.root:
            return

        self.config.parse_file(self.config_path, skip_unreadable=False)

        self._read_category_rules(self.categories_config)
        self._read_file_rules(self.files_config)
        self._read_file_actions(self.actions_config)

    def _read_category_rules(self, rules):
        for rule_text, extra_categories in rules.items():
            rule = self.rule_lexer.get_rule(rule_text)
            self.category_rules.append((rule, helpers.flatten(extra_categories)))

    def _read_file_rules(self, rules):
        for rule_text, filenames in rules.items():
            rule = self.rule_lexer.get_rule(rule_text)
            for filename in helpers.flatten(filenames, ' '):
                self.file_rules.append((rule, filename))

    def _read_file_actions(self, actions):
        for filename, action_text in actions.items():
            action_parts = action_text.strip().split(' ', 1)
            action = action_parts.pop(0)
            if action_parts:
                option_text = action_parts[0]
                options = self.action_lexer.get_options(option_text)
            else:
                options = {}

            self.file_configs[filename] = FileConfig(action, **options)


class Env(object):
    """Holds all configuration for the current environment.

    Attributes:
        repository (Repository): the parsed view of the active repository
        root (str): the path to the repository root
        config (MergedConfig): active configuration directives
    """

    def __init__(self, root, repository, config):
        self.root = root
        self.repository = repository
        self.config = config
        target = self.config.get('target')
        if target:
            target = helpers.get_absolute_path(target, base=self.root)
        self.target = target

        self._forward_fs = self._backward_fs = self._uconf_fs = self._repo_fs = None

    @property
    def uconf_dir(self):
        return os.path.join(self.root, constants.REPO_SUBFOLDER)

    def isset(self, key):
        """Check whether a non-default has been set for a given key."""
        value = self.get(key, default=confutils.NoDefault)
        return value != confutils.NoDefault

    def get(self, key, default=None):
        return self.config.get(key, default)

    def getlist(self, key, default=(), separator=' '):
        value = self.get(key, default=default)
        if isinstance(value, compat.text_types):
            value = value.split(separator)
        return list(value)

    def get_active_repository(self, initial_cats):
        # TODO(rbarrois): consider memoizing
        return self.repository.extract(initial_cats)

    def get_forward_fs(self):
        if self._forward_fs is None:
            self._forward_fs = fs.FSLoader(self.target,
                    dry_run=self.get('dry_run', False),
                    default_encoding=self.get('file_encoding', 'utf8'))
        return self._forward_fs

    def get_backward_fs(self):
        if self._backward_fs is None:
            self._backward_fs = fs.FSLoader(self.root,
                    dry_run=self.get('dry_run', False),
                    default_encoding=self.get('file_encoding', 'utf8'))
        return self._backward_fs

    def get_uconf_fs(self):
        """Retrieve the filesystem associated with the private uconf dir."""
        if self._uconf_fs is None:
            self._uconf_fs = fs.FSLoader(self.uconf_dir,
                    dry_run=self.get('dry_run', False),
                    default_encoding=self.get('file_encoding', 'utf8'))
        return self._uconf_fs

    def get_repo_fs(self):
        """Retrieve a filesystem for the repository, including uconf."""
        if self._repo_fs is None:
            self._repo_fs = fs.FSLoader(self.root,
                    dry_run=self.get('dry_run', False),
                    default_encoding=self.get('file_encoding', 'utf8'))
        return self._repo_fs

    @classmethod
    def _walk_root(cls, base):
        """Walk to the top of a directory tree until a repository root is found.

        Stops at the first folder containing a '.uconf' subdirectory.
        """
        current = helpers.get_absolute_path(base)
        prev = None
        while prev != current:
            maybe_config = os.path.join(current, constants.REPO_SUBFOLDER)
            if os.access(maybe_config, os.F_OK):
                dirmode = os.stat(maybe_config).st_mode
                if stat.S_ISDIR(dirmode):
                    return current
            prev, current = current, os.path.dirname(current)

    @classmethod
    def _read_config(cls, repo_root=None, config_files=constants.CONFIG_FILES):
        """Read all relevant config files."""
        if repo_root:
            repo_root = cls._walk_root(repo_root)

        config = confutils.ConfigFile()

        for config_file in config_files:
            config_file = helpers.get_absolute_path(config_file)
            config.parse_file(config_file, skip_unreadable=True)

        if repo_root:
            repo_config = os.path.join(repo_root,
                constants.REPO_SUBFOLDER, 'config')
            config.parse_file(repo_config, skip_unreadable=False)

        return repo_root, config

    @classmethod
    def _merge_config(cls, config, sections=(), extra=None):
        merged = confutils.MergedConfig()
        if extra is not None:
            merged.add_options(extra)

        for section in sections:
            merged.add_options(config.section_view(section))

        merged.add_options(config.section_view('core'))

        return merged

    @classmethod
    def from_files(cls, repo_root=None, config_files=constants.CONFIG_FILES,
            sections=(), extra=None):
        """Build a Env from basic informations:

        - Path to a repository root
        - List of global configuration files to read
        - List of configuration sections to take into account
        - Dict of extra configuration values
        """

        repo_root, config = cls._read_config(repo_root=repo_root,
                config_files=config_files)

        repo = Repository(root=repo_root)

        config_view = cls._merge_config(config, sections=sections, extra=extra)

        return cls(root=repo_root, config=config_view, repository=repo)
